#define numa_set_interleave_mask(m)     numa_set_interleave_mask_compat(m)
#define numa_get_interleave_mask()      numa_get_interleave_mask_compat()
#define numa_bind(m)                    numa_bind_compat(m)
#define numa_get_membind(m)             numa_get_membind_compat(m)
#define numa_set_membind(m)             numa_set_membind_compat(m)
#define numa_alloc_interleaved_subset(s,m) numa_alloc_interleaved_subset_compat(s,m)
#define numa_run_on_node_mask(m)        numa_run_on_node_mask_compat(m)
#define numa_get_run_node_mask()        numa_get_run_node_mask_compat()
#define numa_interleave_memory(st,si,m) numa_interleave_memory_compat(st,si,m)
#define numa_tonodemask_memory(st,si,m) numa_tonodemask_memory_compat(st,si,m)
#define numa_sched_getaffinity(p,l,m)   numa_sched_getaffinity_compat(p,l,m)
#define numa_sched_setaffinity(p,l,m)   numa_sched_setaffinity_compat(p,l,m)
#define numa_node_to_cpus(n,b,bl)       numa_node_to_cpus_compat(n,b,bl)
#define nodemask_zero(m)		nodemask_zero_compat(m)
#define nodemask_set(m, n)		nodemask_set_compat(m, n)
#define nodemask_clr(m, n)		nodemask_clr_compat(m, n)
#define nodemask_isset(m, n)		nodemask_isset_compat(m, n)
#define nodemask_equal(a, b)		nodemask_equal_compat(a, b)
